@@ -63,6 +63,7 @@ char *err_list[] =
   [ERR_BOOT_FAILURE] = "Unknown boot failure",
   [ERR_BOOT_FEATURES] = "Unsupported Multiboot features requested",
   [ERR_DEV_FORMAT] = "Unrecognized device string",
+  [ERR_DEV_NEED_INIT] = "Device not initialized yet",
   [ERR_DEV_VALUES] = "Invalid device requested",
   [ERR_EXEC_FORMAT] = "Invalid or unsupported executable format",
   [ERR_FILELENGTH] =
@@ -73,7 +74,6 @@ char *err_list[] =
   [ERR_GEOM] = "Selected cylinder exceeds maximum supported by BIOS",
   [ERR_NEED_LX_KERNEL] = "Linux kernel must be loaded before initrd",
   [ERR_NEED_MB_KERNEL] = "Multiboot kernel must be loaded before modules",
-  [ERR_NEED_SERIAL] = "Serial device not configured",
   [ERR_NO_DISK] = "Selected disk does not exist",
   [ERR_NO_DISK_SPACE] = "No spare sectors on the disk",
   [ERR_NO_PART] = "No such partition",
